@@ -2,7 +2,7 @@
 # -*- tcl -*-
 # RSA
 #
-# (c) 2010 Roy Keene.
+# (c) 2010, 2011, 2012, 2013 Roy Keene.
 #	 BSD Licensed.
 
 # # ## ### ##### ######## #############
@@ -186,7 +186,7 @@ namespace eval ::pki {
 
 	variable handlers
 	array set handlers {
-		rsa                            {::pki::rsa::encrypt ::pki::rsa::decrypt ::pki::rsa::generate}
+		rsa                            {::pki::rsa::encrypt ::pki::rsa::decrypt ::pki::rsa::generate ::pki::rsa::serialize_key}
 	}
 
 	variable INT_MAX [expr {[format "%u" -1] / 2}]
@@ -242,6 +242,13 @@ proc ::pki::_powm {x y m} {
 	return $retval
 }
 
+## **NOTE** Requires that "m" be prime
+### a^-1 === a^(m-2)    (all mod m)
+proc ::pki::_modi {a m} {
+	return [_powm $a [expr {$m - 2}] $m]
+	
+}
+
 proc ::pki::_oid_number_to_name {oid} {
 	set oid [join $oid .]
 
@@ -419,6 +426,42 @@ proc ::pki::rsa::decrypt {mode input keylist} {
 	return $retval
 }
 
+proc ::pki::rsa::serialize_key {keylist} {
+	array set key $keylist
+
+	foreach entry [list n e d p q] {
+		if {![info exists key($entry)]} {
+			return -code error "Key does not contain an element $entry"
+		}
+	}
+
+	# Exponent 1
+	## d (mod p-1)
+	set e1 [expr {$key(d) % ($key(p) - 1)}]
+
+	# Exponent 2
+	#set e2 [expr d mod (q-1)]
+	set e2 [expr {$key(d) % ($key(q) - 1)}]
+
+	# Coefficient
+	## Modular multiplicative inverse of q mod p
+	set c [::pki::_modi $key(q) $key(p)]
+
+	set ret [::asn::asnSequence \
+			[::asn::asnBigInteger [::math::bignum::fromstr 0]] \
+			[::asn::asnBigInteger [::math::bignum::fromstr $key(n)]] \
+			[::asn::asnBigInteger [::math::bignum::fromstr $key(e)]] \
+			[::asn::asnBigInteger [::math::bignum::fromstr $key(d)]] \
+			[::asn::asnBigInteger [::math::bignum::fromstr $key(p)]] \
+			[::asn::asnBigInteger [::math::bignum::fromstr $key(q)]] \
+			[::asn::asnBigInteger [::math::bignum::fromstr $e1]] \
+			[::asn::asnBigInteger [::math::bignum::fromstr $e2]] \
+			[::asn::asnBigInteger [::math::bignum::fromstr $c]] \
+	]
+
+	return [list data $ret begin "-----BEGIN RSA PRIVATE KEY-----" end "-----END RSA PRIVATE KEY-----"]
+}
+
 proc ::pki::_lookup_command {action keylist} {
 	array set key $keylist
 
@@ -434,6 +477,9 @@ proc ::pki::_lookup_command {action keylist} {
 		"generate" {
 			set idx 2
 		}
+		"serialize_key" {
+			set idx 3
+		}
 	}
 
 	set cmdlist $::pki::handlers($type)
@@ -673,6 +719,26 @@ proc ::pki::verify {signedmessage checkmessage keylist {algo default}} {
 	return true
 }
 
+proc ::pki::key {keylist {password ""} {encodePem 1}} {
+	set serialize_key [::pki::_lookup_command serialize_key $keylist]
+
+	if {$serialize_key eq ""} {
+		array set key $keylist
+
+		return -code error "Do not know how to serialize an $key(type) key"
+	}
+
+	array set retval_parts [$serialize_key $keylist]
+
+	if {$encodePem} {
+		set retval [::pki::_encode_pem $retval_parts(data) $retval_parts(begin) $retval_parts(end) $password]
+	} else {
+		set retval $retval_parts(data)
+	}
+
+	return $retval
+}
+
 proc ::pki::_parse_init {} {
 	if {[info exists ::pki::_parse_init_done]} {
 		return
@@ -728,9 +794,13 @@ proc ::pki::_getopensslkey {password salt bytes} {
 	return $ret
 }
 
-proc ::pki::_encode_pem {data begin end} {
+proc ::pki::_encode_pem {data begin end {password ""}} {
 	set ret ""
 
+	if {$password != ""} {
+		return -code error "Passwords are currently unsupported !"
+	}
+
 	append ret "${begin}\n"
 	if {$::pki::rsa::base64_binary} {
 		append ret [binary encode base64 -maxlen 64 $data]
@@ -851,11 +921,15 @@ proc ::pki::pkcs::parse_key {key {password ""}} {
 	::asn::asnGetBigInteger key ret(n)
 	::asn::asnGetBigInteger key ret(e)
 	::asn::asnGetBigInteger key ret(d)
+	::asn::asnGetBigInteger key ret(p)
+	::asn::asnGetBigInteger key ret(q)
 
 	set ret(n) [::math::bignum::tostr $ret(n)]
 	set ret(e) [::math::bignum::tostr $ret(e)]
 	set ret(d) [::math::bignum::tostr $ret(d)]
-	set ret(l) [expr {int([::pki::_bits $ret(n)] / 8) * 8}]
+	set ret(p) [::math::bignum::tostr $ret(p)]
+	set ret(q) [::math::bignum::tostr $ret(q)]
+	set ret(l) [expr {int([::pki::_bits $ret(n)] / 8.0000 + 0.5) * 8}]
 	set ret(type) rsa
 
 	return [array get ret]
@@ -1070,7 +1144,7 @@ proc ::pki::x509::parse_cert {cert} {
 
 			set ret(n) [::math::bignum::tostr $ret(n)]
 			set ret(e) [::math::bignum::tostr $ret(e)]
-			set ret(l) [expr {int([::pki::_bits $ret(n)] / 8) * 8}]
+			set ret(l) [expr {int([::pki::_bits $ret(n)] / 8.0000 + 0.5) * 8}]
 			set ret(type) rsa
 		}
 	}
@@ -1645,12 +1719,26 @@ proc ::pki::rsa::generate {bitlength {exponent 0x10001}} {
 
 	# Step 1. Pick 2 numbers that when multiplied together will give a number with the appropriate length
 	set componentbitlen [expr {$bitlength / 2}]
+	set bitmask [expr {(1 << $componentbitlen) - 1}]
 
 	set p 0
 	set q 0
 	while 1 {
 		set plen [::pki::_bits $p]
 		set qlen [::pki::_bits $q]
+
+		if {$plen >= $componentbitlen} {
+			set p [expr {$p & $bitmask}]
+
+			set plen [::pki::_bits $p]
+		}
+
+		if {$qlen >= $componentbitlen} {
+			set q [expr {$q & $bitmask}]
+
+			set qlen [::pki::_bits $q]
+		}
+
 		if {$plen >= $componentbitlen && $qlen >= $componentbitlen} {
 			break
 		}
@@ -1664,20 +1752,10 @@ proc ::pki::rsa::generate {bitlength {exponent 0x10001}} {
 		set xmask [expr {(1 << $xlen) - 1}]
 		set ymask [expr {(1 << $ylen) - 1}]
 
-		set plen [::pki::_bits $p]
-		set qlen [::pki::_bits $q]
-
-		if {$plen < $componentbitlen} {
-			set p [expr {($p << $xlen) + ($x & $xmask)}]
-		}
-		if {$qlen < $componentbitlen} {
-			set q [expr {($q << $ylen) + ($y & $ymask)}]
-		}
+		set p [expr {($p << $xlen) + ($x & $xmask)}]
+		set q [expr {($q << $ylen) + ($y & $ymask)}]
 	}
 
-	set bitmask [expr {(1 << $componentbitlen) - 1}]
-	set p [expr {$p & $bitmask}]
-	set q [expr {$q & $bitmask}]
 
 	# Step 2. Verify that "p" and "q" are useful
 	## Step 2.a. Verify that they are not too close
@@ -1731,6 +1809,10 @@ proc ::pki::rsa::generate {bitlength {exponent 0x10001}} {
 	set retkey(e) $e
 	set retkey(l) $bitlength
 
+	# Step 7. Record additional information that will be needed to write out a PKCS#1 compliant key
+	set retkey(p) $p
+	set retkey(q) $q
+
 	return [array get retkey]
 }
 
@@ -1740,4 +1822,4 @@ proc ::pki::rsa::generate {bitlength {exponent 0x10001}} {
 # # ## ### ##### ######## #############
 ## Ready
 
-package provide pki 0.3
+package provide pki 0.4
